#!/usr/bin/env python3

import inspect
from enum import Enum
from typing import Callable


class ArgumentSatisfactionEnum(Enum):
    none = "none"
    args = "args"
    kws = "kws"
    argskws = "argskws"


def partial_satisfied(partial_fn: Callable) -> bool:
    signature = inspect.signature(partial_fn.func)
    try:
        signature.bind(*partial_fn.args, **partial_fn.keywords)
        return True
    except TypeError:
        return False



#!/usr/bin/env python3

from typing import Sequence

from shapely.geometry import Polygon, LineString


def is_2d_translatable(): ...


def has_touches_across_centroid_buffer(
    poly: Polygon,
    exterior_intersections: Sequence[LineString],
    centroid_buffer_size: float = 0.1,
) -> bool:
    poly.centroid.buffer()

    ...


def is_simple_shape(): ...


def has_3_sides_touching(): ...



from pathlib import Path

# noinspection PyUnresolvedReferences
from qgis.analysis import QgsGcpGeometryTransformer, QgsGcpTransformerInterface

# noinspection PyUnresolvedReferences
from qgis.core import QgsPointXY

__all__ = ["get_gcp_transformer_from_file"]

from .read_gcp_read import read_gcp_file


def get_gcp_transformer_from_file(
    gcp_points_file_path: Path,
    method: QgsGcpTransformerInterface = QgsGcpTransformerInterface.TransformMethod.Helmert,
    *,
    filter_comments: bool = True
) -> QgsGcpGeometryTransformer:
    source_xy, dest_xy = read_gcp_file(
        gcp_points_file_path, filter_comments=filter_comments
    )

    assert (len(source_xy) == len(dest_xy)) and len(
        source_xy
    ) >= QgsGcpTransformerInterface.create(
        QgsGcpTransformerInterface.TransformMethod(method)
    ).minimumGcpCount()

    return QgsGcpGeometryTransformer(
        method, (QgsPointXY(*s) for s in source_xy), (QgsPointXY(*d) for d in dest_xy)
    )



import logging
from enum import Enum
from typing import Collection, Tuple, Union

import shapely
from geojson import LineString
from shapely import affinity

__all__ = [
    "compute_center_principal_axes",
    "buffer_principal_axis",
    "buffer_secondary_axis",
    "PrincipalityMeasure",
    "other_mass_projection_is_longer",
]


class PrincipalityMeasure(Enum):
    double_doors = "double_doors"
    length = "length"
    density = "density"
    hybrid = "hybrid"


logger = logging.getLogger(__name__)


def other_mass_projection_is_longer(
    poly: shapely.geometry.base.BaseGeometry,
    first_axis: LineString,
    other_axis: shapely.LineString,
    hybrid_normalised_axes_projections: bool = False,
) -> bool:
    first_axis_projection_length = first_axis.project(
        poly.centroid, normalized=hybrid_normalised_axes_projections
    )
    other_axis_projection_length = other_axis.project(
        poly.centroid, normalized=hybrid_normalised_axes_projections
    )

    return other_axis_projection_length > first_axis_projection_length


def compute_center_principal_axes(
    poly: shapely.geometry.base.BaseGeometry,
    principality_measure: PrincipalityMeasure = PrincipalityMeasure.length,
    hybrid_ratio_threshold: float = 0.1,
    oppose_hybrid_ratio_cases: bool = True,  # TODO: WACKY
    normalised_projections: bool = False,
    translate_axes: bool = True,
) -> Tuple[shapely.LineString, shapely.LineString]:
    if not isinstance(poly, shapely.geometry.base.BaseGeometry):
        assert isinstance(
            poly, Collection
        ), f"{poly} was an collection of shapely objects"
        poly = shapely.unary_union(poly)

    minimum_rotated_rectangle: shapely.Polygon = poly.minimum_rotated_rectangle

    x, y = minimum_rotated_rectangle.exterior.coords.xy

    first_axis = shapely.LineString(
        [shapely.Point(x[0], y[0]), shapely.Point(x[1], y[1])]
    )
    other_axis = shapely.LineString(
        [shapely.Point(x[1], y[1]), shapely.Point(x[2], y[2])]
    )

    principal_axis = first_axis
    secondary_axis = other_axis

    if principality_measure == PrincipalityMeasure.length:
        if first_axis.length < other_axis.length:
            principal_axis, secondary_axis = secondary_axis, principal_axis

    elif principality_measure == PrincipalityMeasure.density:
        if other_mass_projection_is_longer(
            poly, first_axis, other_axis, normalised_projections
        ):
            principal_axis, secondary_axis = secondary_axis, principal_axis

    elif principality_measure == PrincipalityMeasure.hybrid:
        ratio = first_axis.length / other_axis.length
        if ratio < 1.0 - hybrid_ratio_threshold:
            if oppose_hybrid_ratio_cases:
                principal_axis, secondary_axis = secondary_axis, principal_axis
        elif ratio > 1.0 + hybrid_ratio_threshold:
            if not oppose_hybrid_ratio_cases:
                principal_axis, secondary_axis = secondary_axis, principal_axis
        elif other_mass_projection_is_longer(
            poly, first_axis, other_axis, normalised_projections
        ):
            principal_axis, secondary_axis = secondary_axis, principal_axis
        else:
            ...  # do not swap

    elif principality_measure == PrincipalityMeasure.double_doors:
        ...  # TODO: Grow centroid, test for intersection?

    else:
        raise NotImplementedError(f"{principality_measure=} is not supported")

    if translate_axes:
        return (
            affinity.translate(
                principal_axis,
                xoff=minimum_rotated_rectangle.centroid.x - principal_axis.centroid.x,
                yoff=minimum_rotated_rectangle.centroid.y - principal_axis.centroid.y,
            ),
            affinity.translate(
                secondary_axis,
                xoff=minimum_rotated_rectangle.centroid.x - secondary_axis.centroid.x,
                yoff=minimum_rotated_rectangle.centroid.y - secondary_axis.centroid.y,
            ),
        )

    return principal_axis, secondary_axis


def buffer_principal_axis(
    poly: shapely.geometry.base.BaseGeometry, distance: float = 1.4, **kwargs
) -> Union[shapely.Polygon, shapely.MultiPolygon]:
    pax, sax = compute_center_principal_axes(poly, **kwargs)
    return shapely.buffer(
        pax,
        sax.length / distance,
        # single_sided=True,
        cap_style=shapely.BufferCapStyle.flat,
    )


def buffer_secondary_axis(
    poly: shapely.geometry.base.BaseGeometry, distance: float = 1.4, **kwargs
) -> Union[shapely.Polygon, shapely.MultiPolygon]:
    pax, sax = compute_center_principal_axes(poly, **kwargs)
    return shapely.buffer(
        sax,
        pax.length / distance,
        # single_sided=True,
        cap_style=shapely.BufferCapStyle.flat,
    )


if __name__ == "__main__":

    def juashdu():
        door_wkt = """Polygon ((0.41771353596173888 0.45279106470461222, 0.42202990913524002
    0.44398345695594388, 0.42503883676252752 0.4322059060841496, 0.42406995639706679 0.43195837661635911,
    0.42108655796187611 0.44363600107054563, 0.41687005351423739 0.45223982556683989, 0.41178976975322512
    0.45824659085939767, 0.4062094412971064 0.46209824156605428, 0.40048080898385668 0.46422433579765993,
    0.39496063520860603 0.46503887492821949, 0.39001768047859658 0.46495146513763091, 0.38602765447831972
    0.46437481857520818, 0.38418851989281072 0.46392574885392379, 0.39047171136412068 0.42633994643201001,
    0.38948539794731218 0.42617506512978409, 0.38321620905574583 0.46367710467449691, 0.3824000529003504
    0.46341698788591579, 0.38209639266092799 0.46436976826931692, 0.38304962997471709 0.46467357413683308,
    0.38278232913254839 0.46627255718340038, 0.38376864254935689 0.46643743848562641, 0.38402318794159729
    0.46491475803511939, 0.38583709534983551 0.46535766792364591, 0.38993700301857598 0.46595019480524158,
    0.39502521705723481 0.4660401733196895, 0.40073048270725042 0.46519832262596478, 0.40667459603281331
    0.46299225591362603, 0.41246735498600151 0.4589939815663383, 0.41771353596173888 0.45279106470461222))"""

        buffer_secondary_axis(
            shapely.from_wkt(door_wkt), principality_measure=PrincipalityMeasure.hybrid
        )

    juashdu()



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 02-12-2020
           """

from typing import Tuple, Optional, Any

# noinspection PyUnresolvedReferences
from qgis.PyQt import QtWidgets
from warg import AlsoDecorator, passes_kws_to

from jord.qt_utilities import WindowModalityEnum

__all__ = [
    "make_dialog_progress_bar",
    "DialogProgressBar",
    "make_progress_bar",
    "InjectedProgressBar",
]


def make_progress_bar(
    *,
    progress: int = 0,
    min_value: int = 0,
    max_value: int = 100,
    parent: Optional[Any] = None
) -> QtWidgets.QProgressBar:
    bar = QtWidgets.QProgressBar(parent)
    bar.setTextVisible(True)
    bar.setValue(min_value)
    bar.setValue(progress)
    bar.setMaximum(max_value)
    return bar


def make_dialog_progress_bar(
    *,
    progress: int = 0,
    minimum_width: int = 300,
    min_value: int = 0,
    max_value: int = 100,
    title: str = "Progress",
    label: str = "",
    parent: Optional[Any] = None
) -> Tuple[QtWidgets.QDialog, QtWidgets.QProgressBar]:
    """
    Create a progress bar dialog.

    :param parent:
    :param title:
    :param label:
    :param min_value:
    :param max_value:
    :param progress: The progress to display.
    :type progress: int
    :param minimum_width: The minimum width of the dialog.
    :type minimum_width: int
    :return: The dialog.
    :rtype: Tuple[QtWidgets.QDialog, QtWidgets.QProgressBar]
    """
    dialog = QtWidgets.QProgressDialog(parent)
    dialog.setWindowTitle(title)
    dialog.setLabelText(label)

    bar = make_progress_bar(
        parent=dialog, progress=progress, min_value=min_value, max_value=max_value
    )

    dialog.setBar(bar)
    dialog.setMinimumWidth(minimum_width)
    dialog.setWindowModality(WindowModalityEnum.window.value)

    return dialog, bar


class DialogProgressBar(AlsoDecorator):  # TODO This freezes!
    # TODO Make it formatable with a unit or a title
    # self.progressBar.setFormat(f"{hms} - %p%")
    @passes_kws_to(make_dialog_progress_bar)
    def __init__(self, **kwargs):
        (self._progress_dialog, self._progress_bar) = make_dialog_progress_bar(**kwargs)

    def __enter__(self):
        if self._progress_dialog:
            # self._progress_dialog.forceShow()?
            self._progress_dialog.show()
            # time.sleep(0.2)  # WAIT FOR IT TO BE RENDERED?
            # self._progress_dialog.exec_() # RUN IT?
        return self._progress_bar

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._progress_dialog:
            self._progress_dialog.close()


class InjectedProgressBar(AlsoDecorator):
    @passes_kws_to(make_progress_bar)
    def __init__(self, **kwargs):
        self._parent = kwargs.pop("parent", None)
        self._progress_bar = make_progress_bar(**kwargs)

    def __enter__(self):
        if self._parent:
            self._parent.addWidget(self._progress_bar)
        return self._progress_bar

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._parent:
            self._parent.removeWidget(self._progress_bar)


if __name__ == "__main__":

    def calc(x, y):
        from time import sleep

        dialog, bar = make_dialog_progress_bar(0)
        bar.setValue(0)
        bar.setMaximum(100)
        sum_ = 0
        for i in range(x):
            for j in range(y):
                k = i + j
                sum_ += k
            i += 1
            bar.setValue((float(i) / float(x)) * 100)
            sleep(0.1)
        print(sum_)

    # calc(10000, 2000)



#!/usr/bin/env python3
import re
from pathlib import Path
from typing import List, Sequence, Union

from setuptools import find_packages, setup


def python_version_check(major: int = 3, minor: int = 8) -> None:
    """
      :param major:
    :param minor:
    """
    import sys

    assert sys.version_info.major == major and sys.version_info.minor >= minor, (
        f"This project is utilises language features only present Python {major}.{minor} and greater. "
        f"You are running {sys.version_info}."
    )


python_version_check()


def read_reqs(file: str, path: Path) -> List[str]:
    """description"""

    def readlines_ignore_comments(f):
        """description"""
        return [a_ for a_ in f.readlines() if "#" not in a_ and a_]

    def recursive_flatten_ignore_str(seq: Sequence) -> Sequence:
        """description"""
        if not seq:  # is empty Sequence
            return seq
        if isinstance(seq[0], str):
            return seq
        if isinstance(seq[0], Sequence):
            return (
                *recursive_flatten_ignore_str(seq[0]),
                *recursive_flatten_ignore_str(seq[1:]),
            )
        return *seq[:1], *recursive_flatten_ignore_str(seq[1:])

    def unroll_nested_reqs(req_str: str, base_path: Path):
        """description"""
        if req_str.startswith("-r"):
            with open(base_path / req_str.strip("-r").strip()) as f:
                return [
                    unroll_nested_reqs(req.strip(), base_path)
                    for req in readlines_ignore_comments(f)
                ]
        else:
            return (req_str,)

    requirements_group = []
    with open(str(path / file)) as f:
        requirements = readlines_ignore_comments(f)
        for requirement in requirements:
            requirements_group.extend(
                recursive_flatten_ignore_str(
                    unroll_nested_reqs(requirement.strip(), path)
                )
            )

    req_set = set(requirements_group)
    req_set.discard("")
    return list(req_set)


with open(Path(__file__).parent / "jord" / "__init__.py") as project_init_file:
    str_reg_exp = "['\"]([^'\"]*)['\"]"
    content = project_init_file.read()  # get strings from module
    version = re.search(rf"__version__ = {str_reg_exp}", content, re.M).group(1)
    project_name = re.search(rf"__project__ = {str_reg_exp}", content, re.M).group(1)
    author = re.search(rf"__author__ = {str_reg_exp}", content, re.M).group(1)

__author__ = author


class JordPackage:
    """description"""

    @property
    def test_dependencies(self) -> list:
        return read_reqs(
            "requirements_tests.txt", Path(__file__).parent / "requirements"
        )

    @property
    def setup_dependencies(self) -> list:
        """

        :return: requirements

        """
        return read_reqs(
            "requirements_setup.txt", Path(__file__).parent / "requirements"
        )

    @property
    def package_name(self) -> str:
        """

        Returns:

        """
        return "Jord"

    @property
    def url(self) -> str:
        """

        Returns:

        """
        return "https://github.com/MapsPeople/jord"

    @property
    def download_url(self):
        """

        Returns:

        """
        return self.url + "/releases"

    @property
    def readme_type(self):
        """

        Returns:

        """
        return "text/markdown"

    @property
    def packages(self) -> List[Union[bytes, str]]:
        """

        Returns:

        """
        return find_packages(
            exclude=[
                # 'jord/...'
            ]
        )

    @property
    def author_name(self) -> str:
        """

        Returns:

        """
        return author

    @property
    def author_email(self) -> str:
        """

        Returns:

        """
        return "chen@mapspeople.dk"

    @property
    def maintainer_name(self) -> str:
        """

        Returns:

        """
        return self.author_name

    @property
    def maintainer_email(self) -> str:
        """

        Returns:

        """
        return self.author_email

    @property
    def package_data(self) -> dict:
        """

        Returns:

        """
        emds = [str(p) for p in Path(__file__).parent.rglob(".md")]
        return {"jord": [*emds]}

    @property
    def entry_points(self) -> dict:
        """

        Returns:

        """
        return {
            "console_scripts": [
                # "name_of_executable = module.with:function_to_execute"
            ]
        }

    @property
    def extras(self) -> dict:
        """

        Returns:

        """
        these_extras = {
            # 'ExtraName':['package-name; platform_system == "System(Linux,Windows)"'
        }

        path: Path = Path(__file__).parent / "requirements"

        for file in path.iterdir():
            if file.name.startswith("requirements_"):
                group_name_ = "_".join(file.name.strip(".txt").split("_")[1:])
                these_extras[group_name_] = read_reqs(file.name, path)

        all_dependencies = []

        for group_name in these_extras:
            all_dependencies += these_extras[group_name]
        these_extras["all"] = list(set(all_dependencies))

        return these_extras

    @property
    def requirements(self) -> list:
        """

        Returns:

        """
        requirements_out = []
        with open("requirements.txt") as f:
            requirements = f.readlines()

            for requirement in requirements:
                requirements_out.append(requirement.strip())

        return requirements_out

    @property
    def description(self) -> str:
        """

        Returns:

        """
        return "Geodata toolbox"

    @property
    def readme(self) -> str:
        """

        Returns:

        """
        with open("README.md") as f:
            return f.read()

    @property
    def keyword(self) -> str:
        """

        Returns:

        """
        with open("KEYWORDS.md") as f:
            return f.read()

    @property
    def license(self) -> str:
        """

        Returns:

        """
        return "Apache License, Version 2.0"

    @property
    def classifiers(self) -> List[str]:
        """

        Returns:

        """
        return [
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Intended Audience :: End Users/Desktop",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Natural Language :: English",
            # 'Topic :: Scientific/Engineering :: Artificial Intelligence'
            # 'Topic :: Software Development :: Bug Tracking',
        ]

    @property
    def version(self) -> str:
        """

        Returns:

        """
        return version


if __name__ == "__main__":
    pkg = JordPackage()

    setup(
        name=pkg.package_name,
        version=pkg.version,
        packages=pkg.packages,
        package_data=pkg.package_data,
        author=pkg.author_name,
        author_email=pkg.author_email,
        maintainer=pkg.maintainer_name,
        maintainer_email=pkg.maintainer_email,
        description=pkg.description,
        license=pkg.license,
        keywords=pkg.keyword,
        url=pkg.url,
        download_url=pkg.download_url,
        install_requires=pkg.requirements,
        extras_require=pkg.extras,
        entry_points=pkg.entry_points,
        classifiers=pkg.classifiers,
        long_description_content_type=pkg.readme_type,
        long_description=pkg.readme,
        tests_require=pkg.test_dependencies,
        setup_requires=pkg.setup_dependencies,
        include_package_data=True,
        python_requires=">=3.6",
    )

    """
name: str = ...,
      version: str = ...,
      description: str = ...,
      long_description: str = ...,
      author: str = ...,
      author_email: str = ...,
      maintainer: str = ...,
      maintainer_email: str = ...,
      url: str = ...,
      download_url: str = ...,
      packages: list[str] = ...,
      py_modules: list[str] = ...,
      scripts: list[str] = ...,
      ext_modules: list[Extension] = ...,
      classifiers: list[str] = ...,
      distclass: Type[Distribution] = ...,
      script_name: str = ...,
      script_args: list[str] = ...,
      options: Mapping[str, Any] = ...,
      license: str = ...,
      keywords: list[str] | str = ...,
      platforms: list[str] | str = ...,
      cmdclass: Mapping[str, Type[Command]] = ...,
      data_files: list[tuple[str, list[str]]] = ...,
      package_dir: Mapping[str, str] = ...,
      obsoletes: list[str] = ...,
      provides: list[str] = ...,
      requires: list[str] = ...,
      command_packages: list[str] = ...,
      command_options: Mapping[str, Mapping[str, tuple[Any, Any]]] = ...,
      package_data: Mapping[str, list[str]] = ...,
      include_package_data: bool = ...,
      libraries: list[str] = ...,
      headers: list[str] = ...,
      ext_package: str = ...,
      include_dirs: list[str] = ...,
      password: str = ...,
      fullname: str = ...,
"""



#!/usr/bin/env python3

from jord.qlive_utilities.client import QliveClient

__all__ = ["InterfacedQliveClient"]


class InterfacedQliveClient(QliveClient):
    """
    Useful for making client-side validation of data
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def raster(self): ...

    def geometry(self): ...



#!/usr/bin/env python3


from logging import warning
from typing import Optional

# noinspection PyUnresolvedReferences
from qgis.PyQt import QtCore

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 02-12-2020
           """

__all__ = ["disconnect_signal", "connect_signal", "reconnect_signal"]

IS_DEBUGGING = False


def connect_signal(
    signal: QtCore.pyqtSignal, new_handler: Optional[callable] = None
) -> None:
    """

    :param signal:
    :param new_handler:
    :return:
    """
    if new_handler is not None:  # if new_handler is not None, connect it
        signal.connect(new_handler)
    else:
        if IS_DEBUGGING:
            raise Exception("new_handler is None")
        warning("new_handler is None")


def disconnect_signal(
    signal: QtCore.pyqtSignal, old_handler: Optional[callable] = None
) -> None:
    """

    :param signal:
    :param old_handler:
    :return:
    """
    if signal is not None:
        try:
            if old_handler is not None:  # disconnect old_handler(s)
                while True:
                    # the loop is needed for safely disconnecting a specific handler,
                    # because it may have been connected multple times,
                    # and disconnect(slot) only removes one connection at a time.
                    signal.disconnect(old_handler)
            else:  # disconnect all, only available when old_handler is None and we are debugging, as this is bad practice
                if IS_DEBUGGING:
                    signal.disconnect()
        except TypeError:
            pass


def reconnect_signal(
    signal: QtCore.pyqtSignal,
    new_handler: Optional[callable] = None,
    old_handler: Optional[callable] = None,
) -> None:
    """

    :param signal:
    :type signal: QtCore.pyqtSignal
    :param new_handler:
    :type new_handler: Optional[callable]
    :param old_handler:
    :type old_handler: Optional[callable]
    :return:
    :rtype: None
    """
    disconnect_signal(signal, old_handler)
    connect_signal(signal, new_handler)
